/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <libxml/xmlreader.h>
#include "data.h"
#include "$(appName).h"

struct _attribute {
	char *name;
	char *value;
};

typedef struct _attribute attribute_t;
typedef bool (*xml_weather_parse_cb)(xmlTextReaderPtr xml_reader, bool *no_more_nodes, int attr_count, attribute_t *attribs, char *value, xml_weather_t *weather);

static bool __xml_weather_traverse(xmlTextReaderPtr xml_reader, bool *no_more_nodes, xml_weather_parse_cb parse_cb, xml_weather_t *weather);
static bool __xml_weather_parse_root(xmlTextReaderPtr xml_reader, bool *no_more_nodes, int attr_count, attribute_t *attribs, char *value, xml_weather_t *weather);

/*
 XML data structure

	<current>
		<city id="" name="">
			<coord lon="" lat=""></coord>
			<country>Country code</country>
			<sun rise="" set=""></sun>
		</city>
		<temperature value="" min="" max="" unit=""></temperature>
		<humidity value="" unit=""></humidity>
		<pressure value="" unit=""></pressure>
		<wind>
			<speed value="" name=""></speed>
			<direction value="" code="" name=""></direction>
		</wind>
		<clouds value="" name=""></clouds>
		<visibility value=""></visibility>
		<precipitation value="" mode=""></precipitation>
		<weather number="" value="" icon=""></weather>
		<lastupdate value=""></lastupdate>
	</current>
 */

/*
 * @brief Initialization function for data module
 */
void data_initialize(void)
{
	/*
	 * If you need to initialize managing data,
	 * please use this function.
	 */
}

/*
 * @brief Finalization function for data module
 */
void data_finalize(void)
{
	/*
	 * If you need to finalize managing data,
	 * please use this function.
	 */
}

/*
 * @brief: The functions transforms the date and time referenced by the variable
 * of watch_time_h type into the date_time_t structure.
 * @param[watch_time]: the source of the date and time obtained via the time_tick callback.
 * @param[dt]: reference to the time and date destination structure.
 */
void data_get_date_time_from_watch_time(const watch_time_h watch_time, date_time_t *dt)
{
	watch_time_get_hour24(watch_time, &dt->hour);
	watch_time_get_minute(watch_time, &dt->minute);
	watch_time_get_second(watch_time, &dt->second);
	watch_time_get_year(watch_time, &dt->year);
	watch_time_get_month(watch_time, &dt->month);
	watch_time_get_day(watch_time, &dt->day);
	watch_time_get_day_of_week(watch_time, &dt->day_of_week);
}

/*
 * @brief Function starts the process of XML stream parsing in order to extract all the
 * weather related information.
 * @param[in] xml_buffer The XML data obtained from the weather service.
 * @param[in] buffer_size The size of the 'xml_buffer'.
 * @param[in] weather The resulting structure where all the weather information is stored.
 * @return: The function returns 'true' if the weather information was successfully acquired, otherwise
 * 'false' is returned.
 */
bool data_get_weather(const char *xml_buffer, size_t buffer_size, xml_weather_t **weather)
{
	bool ret;
	bool no_more_nodes;

	xmlTextReaderPtr xml_reader = xmlReaderForMemory(xml_buffer, buffer_size, "", NULL, 0);
	if (!xml_reader) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlReaderForMemory() failed.");
		return false;
	}

	xmlTextReaderNormalization(xml_reader);

	ret = __xml_weather_traverse(xml_reader, &no_more_nodes, __xml_weather_parse_root, *weather);

	xmlFreeTextReader(xml_reader);

	return ret;
}

/*
 * @brief Function frees all the memory allocated during the weather data processing.
 * @param[in] weather The weather structure filled-in during the XML data processing.
 */
void data_dispose_weather(xml_weather_t *weather)
{
	if (!weather)
		return;

	free(weather->city.name);
	free(weather->city.country);
	free(weather->weather.icon);

	free(weather);
}

/*
 * @brief Internal function checks whether two provided strings are the same.
 * @param[in] str1 the first string.
 * @param[in] str2 the second string.
 * @return: The function returns 'true' if the strings are the same, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_same_string(const char *str1, const char *str2)
{
	if (strlen(str1) != strlen(str2))
		return false;

	return (strncmp(str1, str2, strlen(str1)) == 0);
}

/*
 * @brief Internal function gets the current XML node.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] no_more_nodes The value of this argument is set to 'true' if there is no node to obtain.
 * @return: The function returns 'true' if the XML node was successfully obtained, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_get_node(xmlTextReaderPtr xml_reader, bool *no_more_nodes)
{
	*no_more_nodes = false;

	int ret = xmlTextReaderRead(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderRead() failed with error = %d.", ret);
		return false;
	} else if (ret == 0) {
		*no_more_nodes = true;
		return false;
	}

	return true;
}

/*
 * @brief Internal function gets the next XML node from the current one.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] no_more_nodes The value of this argument is set to 'true' if there is no node to obtain.
 * @return: The function returns 'true' if the next XML node was successfully obtained, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_get_node_next(xmlTextReaderPtr xml_reader, bool *no_more_nodes)
{
	*no_more_nodes = false;

	int ret = xmlTextReaderNext(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderNext() failed with error = %d.", ret);
		return false;
	} else if (ret == 0) {
		*no_more_nodes = true;
		return false;
	}

	return true;
}

/*
 * @brief Internal function gets all the attributes from the current XML node.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] attr_count The number of extracted arguments is assigned.
 * @param[in] attribs All the extracted attributes are assigned in the form of array of name-value pairs.
 * @return: The function returns 'true' if all the attributes were successfully obtained, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_extract_attributes(xmlTextReaderPtr xml_reader, int *attr_count, attribute_t **attribs)
{
	*attr_count = 0;
	*attribs = NULL;

	int ret = xmlTextReaderHasAttributes(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderHasAttributes() failed with error = %d.", ret);
		return false;
	}

	if (ret == 0)
		return true;

	ret = xmlTextReaderAttributeCount(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderAttributeCount() failed with error = %d.", ret);
		return false;
	}

	*attr_count = ret;
	*attribs = (attribute_t *)malloc(sizeof(attribute_t) * (*attr_count));
	if (!(*attribs)) {
		*attr_count = 0;
		return false;
	}

	ret = xmlTextReaderMoveToFirstAttribute(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderMoveToFirstAttribute() failed with error = %d.", ret);
		return false;
	}

	int i = 0;
	while (ret == 1) {
		(*attribs)[i].name = strdup((const char *)xmlTextReaderConstName(xml_reader));
		(*attribs)[i].value = strdup((const char *)xmlTextReaderConstValue(xml_reader));
		i++;

		ret = xmlTextReaderMoveToNextAttribute(xml_reader);
		if (ret == -1) {
			xmlTextReaderMoveToElement(xml_reader);
			dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderMoveToNextAttribute() failed with error = %d.", ret);
			return false;
		}
	}

	ret = xmlTextReaderMoveToElement(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderMoveToElement() failed with error = %d.", ret);
		return false;
	}

	return true;
}

/*
 * @brief Internal function responsible for releasing the attributes allocated
 * by the __xml_weather_extract_attributes() function.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array to be released.
 */
static void __xml_weather_free_attributes(int attr_count, attribute_t *attribs)
{
	int i;

	if (!attribs)
		return;

	for (i = 0; i < attr_count; i++) {
		free(attribs[i].name);
		free(attribs[i].value);
	}

	free(attribs);
}

/*
 * @brief Internal function gets the value from the element pointed by the current XML node.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] value The value obtained from the XML node's element.
 * @return: The function returns 'true' if the element's value was successfully obtained, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_extract_value(xmlTextReaderPtr xml_reader, char **value)
{
	*value = NULL;

	xmlNodePtr node = xmlTextReaderCurrentNode(xml_reader);
	if (!node) {
		dlog_print(DLOG_ERROR, LOG_TAG, "node == NULL");
		return false;
	}

	int ret = xmlTextReaderHasValue(xml_reader);
	if (ret == -1) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function xmlTextReaderHasValue() failed with error = %d.", ret);
		return false;
	}

	if (ret == 0)
		return true;

	const char *element_value = (const char *)xmlTextReaderConstValue(xml_reader);

	if (element_value)
		*value = strdup(element_value);

	return true;
}

/*
 * @brief Internal function responsible for parsing the current node. This function obtains
 * all the node's attributes and its values as well as the element's value if it exists. Once the
 * relevant data is obtained, the provided callback function for obtained data parsing is invoked.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] parse_cb The callback parser function to be invoked with obtained data passed as arguments.
 * @param[in] weather The resulting structure containing all the parsed information.
 * @return: The function returns 'true' if the node was successfully parsed, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_parse_node(xmlTextReaderPtr xml_reader, xml_weather_parse_cb parse_cb, xml_weather_t *weather)
{
	int attr_count = 0;
	bool res;
	bool no_more_nodes;
	char *value = NULL;
	attribute_t *attribs = NULL;

	if (parse_cb) {
		__xml_weather_extract_value(xml_reader, &value);
		__xml_weather_extract_attributes(xml_reader, &attr_count, &attribs);

		res = parse_cb(xml_reader, &no_more_nodes, attr_count, attribs, value, weather);

		__xml_weather_free_attributes(attr_count, attribs);
		free(value);

		if (!res && !no_more_nodes)
			return false;
	}

	return true;
}

/*
 * @brief Internal function responsible for traversing the entire XML sub-tree. This function obtains
 * all the node's attributes and its values as well as the element's value if it exists. Once the
 * relevant data is obtained, the provided callback function is invoked.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] no_more_nodes The value of this argument is set to 'true' if there is no more node to obtain.
 * @param[in] parse_cb The callback parser function to be invoked with obtained data passed as arguments.
 * @param[in] weather The resulting structure containing all the parsed information.
 * @return: The function returns 'true' if the XML sub-tree was successfully parsed, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_traverse(xmlTextReaderPtr xml_reader, bool *no_more_nodes, xml_weather_parse_cb parse_cb, xml_weather_t *weather)
{
	int type;

	if (!__xml_weather_get_node(xml_reader, no_more_nodes))
		return *no_more_nodes;

	type = xmlTextReaderNodeType(xml_reader);
	if (type == XML_READER_TYPE_END_ELEMENT)
		return true;

	if (type == XML_READER_TYPE_ELEMENT) {
		if (!__xml_weather_parse_node(xml_reader, parse_cb, weather))
			return false;
	}

	while (__xml_weather_get_node_next(xml_reader, no_more_nodes)) {
		type = xmlTextReaderNodeType(xml_reader);
		if (type == XML_READER_TYPE_END_ELEMENT)
			break;

		if (type != XML_READER_TYPE_ELEMENT)
			continue;

		if (!__xml_weather_parse_node(xml_reader, parse_cb, weather))
			return false;
	}

	return true;
}

/*
 * @brief Internal function gets the named attribute value from the set of attributes.
 * @param[in] attr_name The attribute name which value is to be obtained.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attr_value The attribute value obtained based on its name provided in the 'attr_name' argument.
 * @return: The function returns 'true' if the attribute value was successfully obtained, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_get_attribute_value(const char *attr_name, int attr_count, attribute_t *attribs, char **attr_value)
{
	if (!attr_name)
		return false;

	int i;
	for (i = 0; i < attr_count; i++)
		if (__xml_weather_same_string(attribs[i].name, attr_name)) {
			*attr_value = attribs[i].value;
			return true;
		}

	return false;
}

/*
 * @brief Internal function fills-in the city's related information within the 'weather' structure.
 * This function is invoked by the __xml_weather_parse_current() function as a result of traversing
 * 'current' node of the XML tree.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'city' node.
 * @param[in] weather The resulting structure where the city related information is stored.
 */
static void __xml_weather_get_city(int attr_count, attribute_t *attribs, xml_weather_t *weather)
{
	char *attr_name = NULL;

	if (__xml_weather_get_attribute_value("name", attr_count, attribs, &attr_name))
		if (attr_name && strlen(attr_name) > 0)
			weather->city.name = strdup(attr_name);
}

/*
 * @brief Internal function fills-in the country name within the 'weather' structure.
 * This function is invoked by the __xml_weather_parse_city() function as a result of traversing
 * 'city/country' node of the XML tree.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'city/country' node.
 * @param[in] weather The resulting structure where the country name is stored.
 */
static void __xml_weather_get_city_country(xmlTextReaderPtr xml_reader, xml_weather_t *weather)
{
	char *country = (char *)xmlTextReaderReadString(xml_reader);

	if (country && strlen(country) > 0)
		weather->city.country = strdup(country);

	free(country);
}

/*
 * @brief Internal function fills-in the temperature related information within the 'weather' structure.
 * This function is invoked by the __xml_weather_parse_current() function as a result of traversing
 * 'current' node of the XML tree.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'temperature' node.
 * @param[in] weather The resulting structure where the temperature related information is stored.
 */
static void __xml_weather_get_temperature(int attr_count, attribute_t *attribs, xml_weather_t *weather)
{
	char *attr_value = NULL;

	if (__xml_weather_get_attribute_value("value", attr_count, attribs, &attr_value))
		if (attr_value && strlen(attr_value) > 0)
			weather->temperature.value = atof(attr_value);
}

/*
 * @brief Internal function fills-in the weather identification information within the 'weather' structure.
 * This function is invoked by the __xml_weather_parse_current() function as a result of traversing
 * 'weather' node of the XML tree.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'weather' node.
 * @param[in] weather The resulting structure where the weather identification information is stored.
 */
static void __xml_weather_get_weather(int attr_count, attribute_t *attribs, xml_weather_t *weather)
{
	char *attr_icon = NULL;

	if (__xml_weather_get_attribute_value("icon", attr_count, attribs, &attr_icon))
		if (attr_icon && strlen(attr_icon) > 0)
			weather->weather.icon = strdup(attr_icon);
}

/*
 * @brief Internal function triggers the suitable function responsible for arguments values acquisition
 * within the XML 'city' sub tree. This function is invoked by the __xml_weather_parse_current() function
 * as a result of traversing 'current' node of the XML tree.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] no_more_nodes The value of this argument is set to 'true' if there is no node to obtain.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'city' node.
 * @param[in] value This argument is not used here.
 * @param[in] weather The resulting structure where the city related information is stored.
 * @return: The function returns 'true' if the XML 'city' node was successfully parsed, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_parse_city(xmlTextReaderPtr xml_reader, bool *no_more_nodes, int attr_count, attribute_t *attribs, char *value, xml_weather_t *weather)
{
	const char *node_name = (const char *)xmlTextReaderConstName(xml_reader);
	dlog_print(DLOG_DEBUG, LOG_TAG, "node: city->%s", node_name);

	if (__xml_weather_same_string(node_name, "country"))
		__xml_weather_get_city_country(xml_reader, weather);

	return true;
}

/*
 * @brief Internal function triggers the suitable function responsible for arguments values acquisition
 * within the XML 'current' sub tree. This function is invoked by the __xml_weather_parse_root() function
 * as a result of traversing the entire XML tree.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] no_more_nodes The value of this argument is set to 'true' if there is no node to obtain.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'current' node.
 * @param[in] value This argument is not used here.
 * @param[in] weather The resulting structure where the weather information is stored.
 * @return: The function returns 'true' if the XML 'current' node was successfully parsed, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_parse_current(xmlTextReaderPtr xml_reader, bool *no_more_nodes, int attr_count, attribute_t *attribs, char *value, xml_weather_t *weather)
{
	const char *node_name = (const char *)xmlTextReaderConstName(xml_reader);
	dlog_print(DLOG_DEBUG, LOG_TAG, "node: current->%s", node_name);

	if (__xml_weather_same_string(node_name, "city")) {
		__xml_weather_get_city(attr_count, attribs, weather);

		if (!__xml_weather_traverse(xml_reader, no_more_nodes, __xml_weather_parse_city, weather) && !(*no_more_nodes))
			return false;
	} else if (__xml_weather_same_string(node_name, "temperature")) {
		__xml_weather_get_temperature(attr_count, attribs, weather);
	} else if (__xml_weather_same_string(node_name, "weather")) {
		__xml_weather_get_weather(attr_count, attribs, weather);
	}

	return true;
}

/*
 * @brief Internal function triggers the XML tree parsing starting from its root
 * and it is invoked by the data_xml_weather_get() function.
 * @param[in] xml_reader The XML file reader handle.
 * @param[in] no_more_nodes The value of this argument is set to 'true' if there is no node to obtain.
 * @param[in] attr_count The number of attributes stored in the 'attribs' argument.
 * @param[in] attribs The attributes array of 'current' node.
 * @param[in] value This argument is not used here.
 * @param[in] weather The resulting structure where the weather information is stored.
 * @return: The function returns 'true' if the XML file was successfully parsed, otherwise
 * 'false' is returned.
 */
static bool __xml_weather_parse_root(xmlTextReaderPtr xml_reader, bool *no_more_nodes, int attr_count, attribute_t *attribs, char *value, xml_weather_t *weather)
{
	const char *node_name = (const char *)xmlTextReaderConstName(xml_reader);
	dlog_print(DLOG_DEBUG, LOG_TAG, "node: root->%s", node_name);

	if (__xml_weather_same_string(node_name, "current")) {
		if (!__xml_weather_traverse(xml_reader, no_more_nodes, __xml_weather_parse_current, weather) && !(!no_more_nodes))
			return false;
	} else {
		dlog_print(DLOG_WARN, LOG_TAG, "Not handled node: root->%s node", node_name);
	}

	return true;
}
