/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <locations.h>
#include <Ecore.h>
#include "service/location_query.h"

#define LOCATION_QUERY_INTERVAL 60.0

static struct location_query_info {
	location_manager_h manager_h;
	Ecore_Timer *timer;
	location_query_cb callback;
} s_info = {
	.manager_h = NULL,
	.timer = NULL,
	.callback = NULL,
};

static void _location_service_state_changed_cb(location_service_state_e state, void *user_data);
static Eina_Bool _location_timer_cb(void *data);

/*
 * @brief The location service initialization function.
 * @param[in] callback The callback function to be invoked on successful location data acquisition.
 * @return: This function returns 'true' if the location service was initialized successfully.
 * Otherwise 'false' is returned.
 */
bool location_query_initialize(location_query_cb callback)
{
	int ret;

	s_info.callback = callback;

	ret = location_manager_create(LOCATIONS_METHOD_HYBRID, &s_info.manager_h);
	if (ret != LOCATIONS_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "location_manager_create() failed");
		return false;
	}

	ret = location_manager_set_service_state_changed_cb(s_info.manager_h, _location_service_state_changed_cb, NULL);
	if (ret != LOCATIONS_ERROR_NONE) {
		location_query_finalize();
		dlog_print(DLOG_ERROR, LOG_TAG, "location_manager_set_service_state_changed_cb() failed");
		return false;
	}

	ret = location_manager_start(s_info.manager_h);
	if (ret != LOCATIONS_ERROR_NONE) {
		location_query_finalize();
		dlog_print(DLOG_ERROR, LOG_TAG, "location_manager_start() failed");
		return false;
	}

	return true;
}

/*
 * @brief The location service finalization function.
 */
void location_query_finalize(void)
{
	if (s_info.timer)
		ecore_timer_del(s_info.timer);

	if (s_info.manager_h) {
		location_manager_stop(s_info.manager_h);
		location_manager_unset_service_state_changed_cb(s_info.manager_h);
		location_manager_destroy(s_info.manager_h);
		s_info.manager_h = NULL;
	}
}

/*
 * @brief Internal callback function invoked each time the location service state is changed.
 * @param[in] state The state of the location service. Possible states are: LOCATIONS_SERVICE_ENABLED, LOCATIONS_SERVICE_DISABLED.
 * @param[in] user_data The user data passed to the location_manager_set_service_state_changed_cb() function.
 */
static void _location_service_state_changed_cb(location_service_state_e state, void *user_data)
{
	if (state == LOCATIONS_SERVICE_ENABLED) {
		dlog_print(DLOG_INFO, LOG_TAG, "Location service state: ENABLED");

		if (!s_info.timer)
			s_info.timer = ecore_timer_add(1.0, _location_timer_cb, NULL);
	} else {
		dlog_print(DLOG_INFO, LOG_TAG, "Location service state: DISABLED");

		if (s_info.timer) {
			ecore_timer_del(s_info.timer);
			s_info.timer = NULL;
		}
	}
}

/*
 * @brief Internal callback function invoked when the timer's interval elapses responsible for querying the location data.
 * @param[in] user_data The user data passed to the ecore_timer_add() function.
 * @return: This callback function returns 'EINA_TRUE' to continue with the timer.
 */
static Eina_Bool _location_timer_cb(void *data)
{
	int ret;
	double altitude = 0.0;
	double latitude = 0.0;
	double longitude = 0.0;
	time_t timestamp = {0,};

	ecore_timer_interval_set(s_info.timer, LOCATION_QUERY_INTERVAL);

	ret = location_manager_get_position(s_info.manager_h, &altitude, &latitude, &longitude, &timestamp);
	if (ret == LOCATIONS_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Current location: lat = %.6f, lon = %.6f", latitude, longitude);

		if (s_info.callback)
			s_info.callback(latitude, longitude);
	} else {
		dlog_print(DLOG_ERROR, LOG_TAG, "location_manager_get_position() failed");
	}

	return EINA_TRUE;
}
