/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include "service/weather_poll_thread.h"
#include "service/weather_query.h"
#include "$(appName).h"

#define WEATHER_POLL_INTERVAL_SEC 60.0

typedef struct {
	double latitude;
	double longitude;
} geo_location_t;

static struct weather_poll_info {
	Ecore_Timer *tim;
	Ecore_Thread *thread_h;
	bool pause;
	geo_location_t geo_location;
	xml_weather_t *weather;
	weather_poll_data_ready_cb cb;
} s_info = {
	.tim = NULL,
	.thread_h = NULL,
	.pause = false,
	.geo_location = {0,},
	.weather = NULL,
	.cb = NULL,
};

static Eina_Bool __weather_poll_timer_cb(void *data);
static void __weather_data_read_cb(xml_weather_t *weather);
static void __thread_function(void *data, Ecore_Thread *thread);
static void __thread_end_cb(void *data, Ecore_Thread *thread);
static void __thread_cancelled_cb(void *data, Ecore_Thread *thread);

/*
 * @brief Initializes the weather polling by opening the HTTP channel
 * and timer creation for continuous weather data requesting.
 * @param[in] cb The callback function to be invoked when the weather data is
 * successfully acquired.
 * @return: The function returns 'true' if the weather polling was initialized successfully,
 * otherwise 'false' is returned.
 */
bool weather_poll_thread_init(weather_poll_data_ready_cb cb)
{
	if (!cb || !weather_init())
		return false;

	s_info.cb = cb;

	s_info.tim = ecore_timer_add(1.0, __weather_poll_timer_cb, NULL);

	if (!s_info.tim)
		return false;

	return true;
}

/*
 * @brief Finalize the weather polling.
 */
void weather_poll_thread_deinit(void)
{
	ecore_timer_del(s_info.tim);
	s_info.tim = NULL;
	ecore_thread_cancel(s_info.thread_h);
	s_info.thread_h = NULL;
	weather_deinit();
	s_info.cb = NULL;
	s_info.weather = NULL;
}

/*
 * @brief Sets the geographical coordinates for localized weather query.
 * @param[in] latitude The latitude of the geographical location to query for weather data.
 * @param[in] longitude The longitude of the geographical location to query for weather data.
 */
void weather_poll_thread_set_location(double latitude, double longitude)
{
	s_info.geo_location.latitude = latitude;
	s_info.geo_location.longitude = longitude;
}

/*
 * @brief Internal callback function invoked while the timer's interval elapses which
 * was created within the weather_poll_thread_init() function. This function creates the
 * separate thread where the weather service is requested for the data. This operation,
 * together with response acquisition is time consuming, thus it is executed within
 * a separate thread.
 * @param[in] data The user data passed to the ecore_timer_add() function.
 * @return: The function returns 'EINA_TRUE' if the timer needs to go on,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool __weather_poll_timer_cb(void *data)
{
	ecore_timer_interval_set(s_info.tim, WEATHER_POLL_INTERVAL_SEC);

	if (!s_info.thread_h) {
		s_info.thread_h = ecore_thread_run(__thread_function, __thread_end_cb, __thread_cancelled_cb, NULL);

		if (!s_info.thread_h)
			dlog_print(DLOG_ERROR, LOG_TAG, "Failed to create poll thread");
	}

	return EINA_TRUE;
}

/*
 * @brief Internal callback function invoked within a separate thread which was created
 * by the __weather_poll_timer_cb() function. This function sends a request to the
 * weather web service and waits for its response. The response from web service is passed
 * via the additional callback function assigned internally.
 * @param[in] data The user data passed to the ecore_thread_run() function.
 * @param[in] thread The thread object created in __weather_poll_timer_cb() function.
 */
static void __thread_function(void *data, Ecore_Thread *thread)
{
	if (!weather_query(s_info.geo_location.latitude, s_info.geo_location.longitude, __weather_data_read_cb))
		dlog_print(DLOG_ERROR, LOG_TAG, "Failed to make a query");
}

/*
 * @brief Internal callback function invoked when the thread created in __weather_poll_timer_cb()
 * comes to the end. At this point, the weather information query is assumed to be finished,
 * so all the acquired information is passed to the main thread via the callback function
 * assigned within the weather_poll_thread_init() function.
 * @param[in] data The user data passed to the ecore_thread_run() function.
 * @param[in] thread The thread object created in __weather_poll_timer_cb() function.
 */
static void __thread_end_cb(void *data, Ecore_Thread *thread)
{
	if (s_info.cb)
		s_info.cb(s_info.weather);

	data_dispose_weather(s_info.weather);
	s_info.weather = NULL;
	s_info.thread_h = NULL;
}

/*
 * @brief Internal callback function invoked when the thread is canceled.
 * The previously allocated resources are cleaned here.
 * @param[in] data The user data passed to the ecore_thread_run() function.
 * @param[in] thread The thread object created in __weather_poll_timer_cb() function.
 */
static void __thread_cancelled_cb(void *data, Ecore_Thread *thread)
{
	data_dispose_weather(s_info.weather);
	s_info.weather = NULL;
	s_info.thread_h = NULL;
}

/*
 * @brief Internal callback function invoked when weather web service responds
 * with the weather data acquired.
 * @param[in] weather The structure containing all the acquired weather information.
 * @param[inimage_buffer The structure containing the image data representing the current weather.
 */
static void __weather_data_read_cb(xml_weather_t *weather)
{
	s_info.weather = weather;
}

